import * as THREE from '../build/three.module.js';
import Stats from './jsm/libs/stats.module.js';
import { GPUStatsPanel } from './jsm/utils/GPUStatsPanel.js';
import { GUI } from './jsm/libs/dat.gui.module.js';
import { OrbitControls } from './jsm/controls/OrbitControls.js';
import { Line2 } from './jsm/lines/Line2.js';
import { LineMaterial } from './jsm/lines/LineMaterial.js';
import { LineGeometry } from './jsm/lines/LineGeometry.js';
import { GeometryUtils } from './jsm/utils/GeometryUtils.js';

class App {
    constructor() {
        this._initialize();
    }

    _initialize() {
        this.domWebGL = document.createElement('div');
        document.body.appendChild(this.domWebGL);

        let scene = new THREE.Scene();

        let renderer = new THREE.WebGLRenderer({ antialias: true });
        renderer.setClearColor(0x000000, 0.0);
        this.domWebGL.appendChild(renderer.domElement);

        this.renderer = renderer;
        this.scene = scene;

        this._setupCamera();
        this._setupModel();
        this._setupLights();
        this._setupControls();

        const stats = new Stats();
        this.domWebGL.appendChild(stats.dom);
        this._stats = stats;

        const gpuPanel = new GPUStatsPanel(renderer.getContext());
        stats.addPanel(gpuPanel);
        stats.showPanel(0);
        this._gpuPanel = gpuPanel;

        this._setupGUI();

        window.onresize = this.resize.bind(this);

        this.resize();
    }

    _setupControls() {
        const controls = new OrbitControls(this._camera, this.renderer.domElement);
        controls.minDistance = 10;
        controls.maxDistance = 500;
    }

    _setupModel() {
        const positions = [];
        const colors = [];
        const points = GeometryUtils.hilbert3D(new THREE.Vector3(0,0,0),
            20.0, 
            2, 
            0, 1, 2, 3, 4, 5, 6, 7);
        const spline = new THREE.CatmullRomCurve3(points);
        const divisions = Math.round(12*points.length);
        const point = new THREE.Vector3();
        const color = new THREE.Color();

        for(let i=0,l=divisions; i<l; i++) {
            const t = i / l;
            
            spline.getPoint(t, point);
            positions.push(point.x, point.y, point.z);

            color.setHSL(t, 1.0, 0.5);
            colors.push(color.r, color.g, color.b);
        }

        // Line2 (LineGeometry, LineMaterial)
        const geometry = new LineGeometry();
        geometry.setPositions(positions);
        geometry.setColors(colors);

        const matLine = new LineMaterial({
            color: 0xffffff,
            linewidth: 5,
            vertexColors: true,
            dashed: false,
            alphaToCoverage: true
        });

        const line = new Line2(geometry, matLine);
        line.computeLineDistances();
        line.scale.set(1,1,1);
        this.scene.add(line);

        // THREE.Line (THREE.BufferGeometry, THREE.LineBasicMaterial), gl.LINE_STRIP
        const geo = new THREE.BufferGeometry();
        geo.setAttribute('position', new THREE.Float32BufferAttribute(positions, 3));
        geo.setAttribute('color', new THREE.Float32BufferAttribute(colors, 3));

        const matLineBasic = new THREE.LineBasicMaterial({ vertexColors: true });
        const matLineDashed = new THREE.LineDashedMaterial({ 
            vertexColors: true,
            scale: 2,
            dashSize: 1,
            gapSize: 1
        });

        const line1 = new THREE.Line(geo, matLineBasic);
        line1.computeLineDistances();
        line1.visible = false;
        this.scene.add(line1);

        this._line = line;
        this._matLine = matLine;

        this._line1 = line1;
        this._matLineBasic = matLineBasic;
        this._matLineDashed = matLineDashed;
    }

    _setupLights() {
        this.scene.add(new THREE.AmbientLight(0x222222));

        const light = new THREE.PointLight(0xffffff);
        light.position.copy(this._camera.position);
        this.scene.add(light);
    }

    _setupCamera() {
        const camera = new THREE.PerspectiveCamera(40, window.innerWidth / window.innerHeight, 1, 1000);
        camera.position.set(-40, 0, 60);
        //this.scene.add(camera);

        const camera2 = new THREE.PerspectiveCamera(40, 1, 1, 1000);
        camera2.position.copy(camera.position);

        this._camera = camera;
        this._camera2 = camera2;
    }

    update() {
        this._stats.update();
        //this._controls.update();
    }

    render(time) {
        requestAnimationFrame(this.render.bind(this));
        this.update(time);

        this.renderer.setClearColor(0x000000, 0);
        this.renderer.setViewport(0, 0, window.innerWidth, window.innerHeight);

        this._matLine.resolution.set(window.innerWidth, window.innerHeight);

        this._gpuPanel.startQuery();
        this.renderer.render(this.scene, this._camera);
        this._gpuPanel.endQuery();

        this.renderer.setClearColor(0x222222, 1);
        this.renderer.clearDepth();
        this.renderer.setScissorTest(true);
        this.renderer.setScissor(20, 20, window.innerWidth/4, window.innerHeight/4);
        this.renderer.setViewport(20, 20, window.innerWidth/4, window.innerHeight/4);

        this._camera2.position.copy(this._camera.position);
        this._camera2.quaternion.copy(this._camera.quaternion);

        this._matLine.resolution.set(window.innerWidth/4, window.innerHeight/4);
        this.renderer.render(this.scene, this._camera2);
        this.renderer.setScissorTest(false);
    }

    resize() {
        const aspect = window.innerWidth / window.innerHeight;

        this.renderer.setPixelRatio(window.devicePixelRatio);
        this.renderer.setSize(window.innerWidth, window.innerHeight);

        this._camera.aspect = aspect;
        this._camera.updateProjectionMatrix();

        this._camera2.aspect = (window.innerWidth/4) / (window.innerHeight/4);
        this._camera2.updateProjectionMatrix();

        this.render();
    }

    _setupGUI() {
        const gui = new GUI();
        const param = {
            'line type': 0,
            'width (px)': 5,
            'alphaToCoverage': true,
            'dashed': false,
            'dash scale': 1,
            'dash / gap': 1
        };

        gui.add(param, 'line type', { 'LineGeometry': 0, 'gl.LINE': 1 }).onChange((val) => {
            switch(val) {
                case '0':
                    this._line.visible = true;
                    this._line1.visible = false;
                    break;
                case '1': 
                    this._line.visible = false;
                    this._line1.visible = true;
                    break;
            }
        });

        gui.add(param, 'width (px)', 1, 10).onChange((val) => {
            this._matLine.linewidth = val;
        });

        gui.add(param, 'alphaToCoverage').onChange((val) => {
            this._matLine.alphaToCoverage = val;
        });

        gui.add(param, 'dashed').onChange((val) => {
            this._matLine.dashed = val;
            this._line1.material = val ? this._matLineDashed : this._matLineBasic;
        });

        gui.add(param, 'dash scale', 0.5, 2, 0.1).onChange((value) => {
            this._matLine.dashScale = value;
            this._matLineDashed.scale = value;
        });

        gui.add(param, 'dash / gap', {'2 : 1': 0, '1 : 1': 1, '1 : 2': 2}).onChange((val) => {
            switch(val) {
                case '0':
                    this._matLine.dashSize = 2;
                    this._matLine.gapSize = 1;
                    this._matLineDashed.dashSize = 2;
                    this._matLineDashed.gapSize = 1;
                    break;
                case '1':
                    this._matLine.dashSize = 1;
                    this._matLine.gapSize = 1;
                    this._matLineDashed.dashSize = 1;
                    this._matLineDashed.gapSize = 1;
                    break;
                case '2':
                    this._matLine.dashSize = 1;
                    this._matLine.gapSize = 2;
                    this._matLineDashed.dashSize = 1;
                    this._matLineDashed.gapSize = 2;
                    break;
            }
        });
    }
}

window.onload = function () {
    (new App()).render();
}